<?php
/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

/**
 * Cron Job that:
 * - reads scheduled task from esb_schedule table;
 * - instantiates object taking class/interface from task;
 * - converts object into configured output format;
 * - calls endpoint;
 * - in case of error schedules again the task.
 */

namespace FiloBlu\Esb\Cron;

use Exception;
use FiloBlu\Esb\Helper\Format\FormatFactory;
use FiloBlu\Esb\Helper\LoggerProvider;
use FiloBlu\Esb\Model\Schedule;
use FiloBlu\Esb\Model\ScheduleFactory;
use Magento\Catalog\Model\ProductFactory;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\HTTP\ZendClientFactory;
use Magento\Framework\ObjectManagerInterface;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Framework\Webapi\ServiceOutputProcessor;
use Magento\Sales\Api\Data\ItemExtensionFactory;
use Magento\Sales\Api\Data\OrderItemExtensionFactory;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Api\ShipmentRepositoryInterface;
use Magento\Sales\Model\EntityInterface;
use Magento\Sales\Model\Order\ItemFactory;
use Magento\Store\Model\ScopeInterface;
use Monolog\Logger;
use Zend_Db_Expr;
use Zend_Http_Client;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\OrderStatusHistoryInterface;
use Magento\Sales\Api\OrderManagementInterface;
use Magento\Sales\Model\ResourceModel\Order\Tax\Item;

class SendMessage
{
    /**
     * @var ScheduleFactory
     */
    protected $_scheduleFactory;

    /**
     * @var FormatFactory
     */
    protected $_formatFactory;

    /**
     * @var Logger
     */
    protected $_logger;

    /**
     * @var ObjectManagerInterface
     */
    protected $_objectManager;

    /**
     * @var \Magento\Sales\Model\ProductFactory
     */
    protected $_productFactory;

    /**
     * @var ItemFactory
     */
    protected $_orderItemFactory;

    /**
     * @var ItemExtensionFactory
     */
    protected $_itemExtensionFactory;

    /**
     * @var ZendClientFactory
     */
    protected $_httpClientFactory;

    /**
     * @var ServiceOutputProcessor
     */
    protected $_serviceOutputProcessor;

    /**
     * @var ScopeConfigInterface
     */
    protected $_scopeConfig;

    /**
     * @var string
     */
    protected $_response;

    /**
     * @var SerializerInterface
     */
    protected $_serializer;

    /**
     * @param ScheduleFactory $scheduleFactory
     * @param FormatFactory $formatFactory
     * @param ZendClientFactory $httpClientFactory
     * @param LoggerProvider $loggerProvider
     * @param ObjectManagerInterface $objectManager
     * @param ProductFactory $productFactory
     * @param ItemFactory $orderItemFactory
     * @param OrderItemExtensionFactory $itemExtensionFactory
     * @param ServiceOutputProcessor $serviceOutputProcessor
     * @param ScopeConfigInterface $scopeConfig
     * @param SerializerInterface $serializer
     */
    public function __construct(
        ScheduleFactory $scheduleFactory,
        FormatFactory $formatFactory,
        ZendClientFactory $httpClientFactory,
        LoggerProvider $loggerProvider,
        ObjectManagerInterface $objectManager,
        ProductFactory $productFactory,
        ItemFactory $orderItemFactory,
        OrderItemExtensionFactory $itemExtensionFactory,
        ServiceOutputProcessor $serviceOutputProcessor,
        ScopeConfigInterface $scopeConfig,
        SerializerInterface $serializer
    )
    {
        $this->_scheduleFactory = $scheduleFactory;
        $this->_formatFactory = $formatFactory;
        $this->_logger = $loggerProvider->getLogger();
        $this->_objectManager = $objectManager;
        $this->_productFactory = $productFactory;
        $this->_orderItemFactory = $orderItemFactory;
        $this->_itemExtensionFactory = $itemExtensionFactory;
        $this->_httpClientFactory = $httpClientFactory;
        $this->_serviceOutputProcessor = $serviceOutputProcessor;
        $this->_scopeConfig = $scopeConfig;
        $this->_serializer = $serializer;
    }

    public function execute()
    {

        /** Retrieve the pending task to be executed */
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
            ->addFieldToFilter('status', 'pending')
            ->addFieldToFilter('executable_after', ['lteq' => new Zend_Db_Expr('NOW()')])
            ->addFieldToFilter('sendmessage_class', '\\FiloBlu\\Esb\\Cron\\SendMessage')
            ->setOrder('priority', 'ASC');

        /** Instantiate format helper (JSON/XML) */
        $formatData = $this->_formatFactory->create();

        foreach ($scheduleCollection as $schedule) {
            try {
                /** Set schedule status to running to prevent concurrent executions.*/
                $schedule->setStatus('running');
                $schedule->setExecutedAt(new Zend_Db_Expr('NOW()'));
                $schedule->save();

                /** Schedule validation.*/
                $validation = $this->validateSchedule($schedule);
                if (!$validation['result']) {
                    $schedule->setStatus('skipped');
                    $schedule->setResponse($validation['reason']);
                    $schedule->save();
                    continue;
                }


                $message = "";

                /** The scheduled task contains a Repository Interface. Use Repository design pattern to load object. */
                if ($schedule->getObjectRepositoryInterface() !== null) {
                    $objectInterface = $schedule->getObjectRepositoryInterface();
                    $objectRepository = $this->_objectManager->create($objectInterface);
                    $outputObject = $objectRepository->get($schedule->getObjectId());

                    // Add gift wrap product to order
                    if ($objectInterface === '\Magento\Sales\Api\OrderRepositoryInterface') {
                        $this->addGiftWrapProduct($outputObject);
                    }
                    if ($objectInterface === '\Magento\Sales\Api\ShipmentRepositoryInterface') {
                        $this->addExternalInvoices($outputObject);
                    }
                    /* Use serviceOutputProcessor to build output data. See function processApiRequest() of Magento\Webapi\Controller\Rest. */
                    $message = $this->_serviceOutputProcessor->process(
                        $outputObject,
                        $objectInterface,
                        'get'
                    );
                } /** The scheduled task contains an Object Class. Use standard magento loading. */
                elseif ($schedule->getObjectClass() !== null) {
                    $outputObject = $this->_objectManager->create($schedule->getObjectClass())->load($schedule->getObjectId());
                    $message = $outputObject->getData();
                } /** Cannot find interfaces nor classes: cannot instantiate object. */
                else {
                    $schedule->setStatus('permanent_error');
                    $schedule->setResponse('Cannot instantiate Object');
                    $schedule->save();
                    continue;
                }

                //$this->_logger->info("MESSAGE: ".$message);
                $store_id = $outputObject->getStoreId();
                $envelope = [];
                $envelope['platform'] = $this->_scopeConfig->getValue(
                    'esb/general/platform_code',
                    ScopeInterface::SCOPE_STORE,
                    $store_id
                );
                $envelope['market'] = $this->_scopeConfig->getValue(
                    'esb/general/market_code',
                    ScopeInterface::SCOPE_STORE,
                    $store_id
                );
                $envelope['channeltype'] = $this->_scopeConfig->getValue(
                    'esb/general/channeltype',
                    ScopeInterface::SCOPE_STORE,
                    $store_id
                );
                $envelope['message'] = $message;

                $error = false;

                $outputData = $formatData->format($envelope);
                //$this->_logger->info("PAYLOAD: ".$outputData);

                /** Create HTTP Client. Warning: Zend_HTTP_Client is ZF1 not ZF2.*/
                if (strlen($schedule->getEndpoint()) > 0) {
                    $client = $this->_httpClientFactory->create();
                    $client->setUri($schedule->getEndpoint());
                    $client->setRawData($outputData);
                    $client->setEncType($formatData->getContentType());
                    $client->setMethod(Zend_Http_Client::POST);
                    //$this->_logger->info("Client: ".print_r($client, true));
                    $this->_response = $client->request();
                    //$this->_logger->info("Response: ".print_r($this->_response, true));
                    $schedule->setResponse($this->_response->getBody());
                    $schedule->setFinishedAt(new Zend_Db_Expr('NOW()'));
                    if ($this->_response->isSuccessful()) {
                        $schedule->setStatus('success');
                    } else {
                        $this->_logger->error($this->_response->getBody());
                        $schedule->setResponse($this->_response->getBody());
                        $schedule->setStatus('error');
                        $error = true;
                    }
                } else {
                    $schedule->setStatus('no_endpoint');
                }
                if (!$error && $schedule->getObjectAfterMethod() !== null && $schedule->getObjectAfterMethod() !== '') {
                    $methodName = $schedule->getObjectAfterMethod();
                    $methodArgument = $schedule->getObjectAfterMethodArgument();
                    $this->{$methodName}($outputObject, $methodArgument);
                }
                $schedule->save();
            } catch (Exception $e) {
                $this->_logger->error($e->getMessage());
                $schedule->setStatus('error');
                $schedule->setResponse($e->getMessage());
                $schedule->save();
                $error = true;
            }
            if ($error) {
                $this->scheduleAgain($schedule);
            }
        }
        return $this;
    }

    /**
     * Validate schedule for processing
     *
     * @param Schedule $schedule
     * @return array
     */
    protected function validateSchedule($schedule)
    {
        $output = ['result' => true, 'reason' => ''];
        switch ($schedule->getObjectRepositoryInterface()) {
            case "\\Magento\\Sales\\Api\\OrderRepositoryInterface":
                $objectRepository = $this->_objectManager->create(OrderRepositoryInterface::class);
                $order = $objectRepository->get($schedule->getObjectId());
                $skip = false;
                if ($order->getStatus() === 'closed') {
                    $skip = true;
                }
                if ($order->getStatus() === 'blocked') {
                    $skip = true;
                }
                if ($order->getStatus() === 'canceled' && $schedule->getObjectAfterMethodArgument() !== 'canceled') {
                    $skip = true;
                }
                if ($order->getStatus() === 'holded' && $schedule->getObjectAfterMethodArgument() !== 'holded') {
                    $skip = true;
                }
                if ($skip) {
                    $output['result'] = false;
                    $output['reason'] .= 'Order Status is ' . $order->getStatus();
                }
                break;
        }
        return $output;
    }

    /**
     * If gift wrap is selected, add linked product to order, see config esb/gwmapping/gwproducts
     *
     * @param OrderRepositoryInterface $order
     */
    protected function addGiftWrapProduct($order)
    {

        $products = $this->_serializer->unserialize($this->_scopeConfig->getValue(
            'esb/gwmapping/gwproducts',
            ScopeInterface::SCOPE_STORE,
            $order->getStoreId()
        ));

        if (empty($products)) {
            return;
        }

        if ($order->getGwId()) {
            $productSku = null;
            foreach ($products as $productData) {
                if ($productData['wrapping_id'] === '*' || $productData['wrapping_id'] == $order->getGwId()) {
                    $productSku = $productData['product_sku'];
                    break;
                }
            }

            try {

                $product = $this->_productFactory->create()->loadByAttribute('sku', $productSku);
                $giftOrderItem = $this->_orderItemFactory->create();

                $giftOrderItem->setProductId($product->getId());
                $giftOrderItem->setSku($product->getSku());
                $giftOrderItem->setName($product->getName());
                $giftOrderItem->setQtyOrdered(1);
                $giftOrderItem->setQtyInvoiced(1);
                $giftOrderItem->setProductType($product->getTypeId());
                $giftOrderItem->setBasePrice($order->getGwBasePrice());
                $giftOrderItem->setPrice($order->getGwPrice());
                $giftOrderItem->setBasePriceInclTax($order->getGwBasePriceInclTax());
                $giftOrderItem->setPriceInclTax($order->getGwPriceInclTax());
                $giftOrderItem->setBaseTaxAmount($order->getGwBaseTaxAmount());
                $giftOrderItem->setTaxAmount($order->getGwTaxAmount());
                $giftOrderItem->setTaxInvoiced($order->getGwTaxAmount());
                $giftOrderItem->setBaseRowInvoiced($order->getGwBasePrice());
                $giftOrderItem->setBaseRowTotal($order->getGwBasePrice());
                $giftOrderItem->setRowTotal($order->getGwBasePrice());
                $giftOrderItem->setBaseRowTotalInclTax($order->getGwBasePriceInclTax());

                $itemTax = $this->_objectManager->create(Item::class)->getTaxItemsByOrderId($order->getId());
                $percent = 0;
                foreach ($itemTax as $tax) {
                    if ($tax['taxable_item_type'] === 'quote_gw') {
                        $percent = $tax['tax_percent'];
                        break;
                    }
                }
                $giftOrderItem->setTaxPercent($percent);
                $giftOrderItem->setBaseAmountRefunded('0.0000');
                $giftOrderItem->setBaseDiscountAmount('0.0000');
                $giftOrderItem->setBaseDiscountTaxCompensationAmount('0.0000');
                $giftOrderItem->setBaseDiscountTaxCompensationInvoiced('0.0000');
                $giftOrderItem->setDiscountAmount('0.0000');
                $giftOrderItem->setDiscountInvoiced('0.0000');
                $giftOrderItem->setDiscountPercent('0.0000');

                $extensionAttributes = $this->_itemExtensionFactory->create();
                $extensionAttributes->setData('kpi_full_price', $order->getGwBasePriceInclTax());
                $giftOrderItem->setExtensionAttributes($extensionAttributes);

                $order->addItem($giftOrderItem);

            } catch (Exception $e) {
                $this->_logger->error($e->getMessage());
            }

        }

        $orderItems = $order->getItems();
        foreach ($orderItems as $orderItem) {
            if ($orderItem->getGwId()) {
                $productSku = null;
                foreach ($products as $productData) {
                    if ($productData['wrapping_id'] === '*' || $productData['wrapping_id'] == $orderItem->getGwId()) {
                        $productSku = $productData['product_sku'];
                        break;
                    }
                }

                $product = $this->_productFactory->create()->loadByAttribute('sku', $productSku);
                $giftOrderItem = $this->_orderItemFactory->create();

                $giftOrderItem->setProductId($product->getId());
                $giftOrderItem->setSku($product->getSku());
                $giftOrderItem->setName($product->getName() . ' (pack for ' . $orderItem->getSku() . ')');
                $giftOrderItem->setQtyOrdered(1);
                $giftOrderItem->setQtyInvoiced(1);
                $giftOrderItem->setProductType($product->getTypeId());
                $giftOrderItem->setBasePrice($orderItem->getGwBasePrice());
                $giftOrderItem->setPrice($orderItem->getGwPrice());
                $giftOrderItem->setBasePriceInclTax($orderItem->getGwBasePrice() + $orderItem->getGwBaseTaxAmount());
                $giftOrderItem->setPriceInclTax($orderItem->getGwPrice() + $orderItem->getGwTaxAmount());
                $giftOrderItem->setBaseTaxAmount($orderItem->getGwBaseTaxAmount());
                $giftOrderItem->setTaxAmount($orderItem->getGwTaxAmount());
                $giftOrderItem->setTaxInvoiced($orderItem->getGwTaxAmount());
                $giftOrderItem->setBaseRowInvoiced($orderItem->getGwBasePrice());
                $giftOrderItem->setBaseRowTotal($orderItem->getGwBasePrice());
                $giftOrderItem->setRowTotal($orderItem->getGwBasePrice());
                $giftOrderItem->setBaseRowTotalInclTax($orderItem->getGwBasePrice() + $orderItem->getGwBaseTaxAmount());

                $itemTax = $this->_objectManager->create(Item::class)->getTaxItemsByOrderId($order->getId());
                $percent = 0;
                foreach ($itemTax as $tax) {
                    if ($tax['taxable_item_type'] === 'item_gw' && (int)$tax['associated_item_id'] === $orderItem->getId()) {
                        $percent = $tax['tax_percent'];
                        break;
                    }
                }
                $giftOrderItem->setTaxPercent($percent);
                $giftOrderItem->setBaseAmountRefunded('0.0000');
                $giftOrderItem->setBaseDiscountAmount('0.0000');
                $giftOrderItem->setBaseDiscountTaxCompensationAmount('0.0000');
                $giftOrderItem->setBaseDiscountTaxCompensationInvoiced('0.0000');
                $giftOrderItem->setDiscountAmount('0.0000');
                $giftOrderItem->setDiscountInvoiced('0.0000');
                $giftOrderItem->setDiscountPercent('0.0000');

                $extensionAttributes = $this->_itemExtensionFactory->create();
                $extensionAttributes->setData('kpi_full_price', $orderItem->getGwBasePriceInclTax());
                $extensionAttributes->setData('packed_sku', $orderItem->getSku());
                $giftOrderItem->setExtensionAttributes($extensionAttributes);

                $order->addItem($giftOrderItem);
            }
        }
    }

    /**
     * Add External Invoices
     *
     * @param ShipmentRepositoryInterface $shipment
     */
    protected function addExternalInvoices($shipment)
    {
        $order = $shipment->getOrder();
        $orderExtensionAttributes = $order->getExtensionAttributes();

        if ($orderExtensionAttributes && $orderExtensionAttributes->getFbOad()) {
            $additionalDataObject = $orderExtensionAttributes->getFbOad();
            $additionalData = array();
            foreach ($additionalDataObject->getCustomAttributes() as $customAttribute) {
                $additionalData[$customAttribute->getAttributeCode()] = $customAttribute->getValue();
            }
            if (array_key_exists('external_invoices', $additionalData)) {
                $externalInvoices = $additionalData['external_invoices'];
                $invoice = current($externalInvoices);
                $shipmentExtensionAttributes = $shipment->getExtensionAttributes();
                if ($shipmentExtensionAttributes !== null) {
                    $shipmentExtensionAttributes->setDocumentDate($invoice['created_at']);
                    $shipmentExtensionAttributes->setDocumentType($invoice['document_type']);
                    $shipmentExtensionAttributes->setDocumentNumber($invoice['increment_id']);
                    $shipment->setExtensionAttributes($shipmentExtensionAttributes);
                }
            }
        }
    }

    /**
     * Schedule job again. The schedule time depends on the number of previous errors on the same item.
     *
     * @param Schedule $schedule
     * @return Schedule $newSchedule
     * @throws Exception
     */
    protected function scheduleAgain($schedule)
    {
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
            ->addFieldToFilter('status', 'error')
            ->addFieldToFilter('object_id', $schedule->getObjectId())
            ->addFieldToFilter('endpoint', $schedule->getEndpoint());

        if ($schedule->getObjectClass()) {
            $scheduleCollection->addFieldToFilter('object_class', $schedule->getObjectClass());
        }

        if ($schedule->getObjectRepositoryInterface()) {
            $scheduleCollection->addFieldToFilter('object_repository_interface', $schedule->getObjectRepositoryInterface());
        }

        $errors = $scheduleCollection->getSize();
        if ($errors < 5) {
            $newSchedule = $this->_scheduleFactory->create();
            $minutes = 5 * pow(2, $errors - 1);
            $newSchedule->setData(
                [
                    'object_class' => $schedule->getObjectClass(),
                    'object_repository_interface' => $schedule->getObjectRepositoryInterface(),
                    'object_id' => $schedule->getObjectId(),
                    'endpoint' => $schedule->getEndpoint(),
                    'event' => $schedule->getEvent(),
                    'sendmessage_class' => "\FiloBlu\Esb\Cron\SendMessage",
                    'object_after_method' => $schedule->getObjectAfterMethod(),
                    'object_after_method_argument' => $schedule->getObjectAfterMethodArgument(),
                    'status' => 'pending',
                    'priority' => $schedule->getPriority(),
                    'executable_after' => new Zend_Db_Expr('TIMESTAMPADD(MINUTE,' . $minutes . ', NOW())')
                ]
            );
            $newSchedule->save();
            return $newSchedule;
        }

        $schedule->setStatus('permanent_error');
        $schedule->save();
        $this->_logger->error("Won't schedule again. More than 4 errors while scheduling. " . json_encode($schedule));
        return false;
    }

    /**
     * Object After Method: update Order Status
     *
     * @param EntityInterface $object
     * @param $status
     * @return void $newSchedule
     * @throws Exception
     */
    protected function updateOrderStatus($object, $status)
    {
        $orderManagement = $this->_objectManager->get(OrderManagementInterface::class);
        $statusHistory = $this->_objectManager->create(OrderStatusHistoryInterface::class);
        if (is_subclass_of($object, OrderInterface::class)) {
            $order = $object;
        } elseif ($object->getOrderId()) {
            $objectRepository = $this->_objectManager->create(OrderRepositoryInterface::class);
            $order = $objectRepository->get($object->getOrderId());
        } else {
            throw new Exception('Cannot find order to update.');
        }
        $status_to_set = $order->getStatus();
        if ($status !== '*') {
            $status_to_set = $status;
        }
        $statusHistory->setStatus($status_to_set);

        $comment = 'ESB Module. ';
        if ($this->_response) {
            $comment .= $this->parseESBResponse($this->_response->getBody());
        }
        $statusHistory->setComment($comment);
        $orderManagement->addComment($order->getId(), $statusHistory);
    }

    /**
     * Parse JSON response
     *
     * @param mixed $message
     * @return string $message
     */
    protected function parseESBResponse($message)
    {
        $decoded = json_decode($message, true);
        if ($decoded) {
            $output = '';
            if (array_key_exists('flow', $decoded)) {
                $output .= 'Flow: ' . $decoded['flow'] . '<br>';
            }
            if (array_key_exists('message', $decoded)) {
                $output .= $decoded['message'] . '. ';
            }
            if (array_key_exists('description', $decoded)) {
                $output .= $decoded['description'] . '. ';
            }
            if ($output) {
                return $output;
            }
        }
        return $message;
    }
}